/*****************************************************************************\
 *
 * Author: Michal Tth (xtothm00 _AT stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, author will kindly welcome
 * if you give him credit and keep his names with their source code
 * and inform him about any succesful project using this code.
 *
\*****************************************************************************/
#include "gamecontrol.h"

GameControl::GameControl(QObject *parent) : QObject(parent) {
    // load controls, settings
    cont = Controls::instance();
    set = Settings::instance();
}

GameControl::~GameControl() {
    for (int i = 0; i < players.count(); i++) {
        delete players.at(i);
    }
    for (int i = 0; i < projectiles.count(); i++) {
        delete projectiles.at(i);
    }
    for (int i = 0; i < mapObjects.count(); i++) {
        delete mapObjects.at(i);
    }
    for (int i = 0; i < bonuses.count(); i++) {
        delete bonuses.at(i);
    }
}

void GameControl::init_shared() {

    // random generator init
    srand(QTime::currentTime().msec());

    // show grid
    gs->setGridVisible(true);

    if (DEBUG) {
        lab = new QLabel;
        lab->setObjectName("infolab");
        lab->setGeometry(20, 20, 180, 150);
        gs->addToHUD(lab);
    }

    // load kamera
    camera = new GameCamera(gs->getCamera());

    // load selected map
    Maps::instance()->getMapList()->at(set->getSelectedMap())->loadMap(this);

    // load players
    int skip = 0;
    for (int i = 0; i < MAX_PLAYERS; i++) {
        int type = set->playerType[i];
        if (type > 0) {
            Player* p = Player::newPlayer(type, gs);
            p->setName(set->getPlayerName(i));
            spawn(p);
            players.append(p);
            gs->getPortrait(i - skip)->setName(p->getName());
            ((QLabel*) (gs->getPortrait(i - skip)->parent()))->show();
            gs->addModel(p->getModel());
        } else {
            skip++;
        }
    }

    sputnik = new PlayerSputnik();
    spawn(sputnik);
    players.append(sputnik);
    gs->addModel(sputnik->getModel());


}

void GameControl::init() {
    init_shared();
    // set camera mode
    camera->follow(players.at(0));
    camera->switchMode();

    // cheats
    connect(cont, SIGNAL(cheat1()), this, SLOT(slotCheat1()));
    connect(cont, SIGNAL(cheat2()), this, SLOT(slotCheat2()));
    connect(cont, SIGNAL(cheat3()), this, SLOT(slotCheat3()));
    connect(cont, SIGNAL(cheat4()), this, SLOT(slotCheat4()));
    std::cout << "starting gc\n";
}

void GameControl::start() {
    std::cout << "started gc\n";
    timer.start(10);
    connect(&timer, SIGNAL(timeout()), this, SLOT(gameLoop()));
    time.start();
    fps.start();
    lps = 0;
    loopTime.start();
    frames = 0;
}

void GameControl::gameLoop() {
    if (DEBUG_LOOP)printf("gcc gameLoop\n");
    frames++;
    int elapsed = loopTime.elapsed();
    loopTime.restart();


    for(int i=0;i<players.count();i++){
        Player* p = players.at(i);
        Settings * set= Settings::instance();
        if(p->getScore()>=set->score){
            set->resetScores();
            for(int i=0;i<players.count();i++){
                set->playerScore[i]=players.at(i)->getScore();
                set->playerName[i]=players.at(i)->getName();
            }
            emit gameOver();
        }
    }


    // *********************
    // players
    if (DEBUG_LOOP)printf("up pl\n");
    updatePlayers(elapsed);


    // *********************
    // bonuses
    if (DEBUG_LOOP)printf("up bon\n");
    updateBonuses(elapsed);

    // *********************
    // camera
    if (DEBUG_LOOP)printf("up cam\n");
    updateCamera(elapsed);

    // *********************
    // projectiles
    if (DEBUG_LOOP)printf("up pr\n");
    updateProjectiles(elapsed);

    // *********************
    // collisions
    if (DEBUG_LOOP)printf("up col\n");
    updateCollisions(elapsed);

    // *********************
    // objects
    if (DEBUG_LOOP)printf("up ob\n");
    updateObjects(elapsed);


    // >>>>
    // debug output
    if (DEBUG_LOOP)printf("up debug\n");

    if (DEBUG) {
        QString str;

        if (frames && frames % 100 == 0) {
            lps = 100000 / fps.elapsed();
            fps.restart();
        }

        str += QString("Debug window:\n lps: %1\n")
                .arg(lps);
        /*
        if(!players.empty()){
            str+=QString("time:%5\nkrava:\n %1 deg = %2\n x=%3 y=%4\n speed=%6 %7 %8\n")
                    .arg(players.at(0)->getYRotation())
                    .arg(osg::DegreesToRadians(players.at(0)->getYRotation()))
                    .arg((int)players.at(0)->getX())
                    .arg((int)players.at(0)->getY())
                    .arg(time)
                    .arg(players.at(0)->getSpeed().x())
                    .arg(players.at(0)->getSpeed().y())
                    .arg(players.at(0)->getSpeed().z());
        }*/
        /*
        str += QString(" camera:\n focus x=%1 y=%2 \n distance=%3 angle%4 \n eye: x=%5 y=%6 z=%7 \n")
                .arg(camera->getFocus().x())
                .arg(camera->getFocus().y())
                .arg(camera->getDistance())
                .arg(camera->getAngle())
                .arg(camera->getEye().x())
                .arg(camera->getEye().y())
                .arg(camera->getEye().z());
        if (camera->getMode() == GameCamera::FOCUS) {
            str += " mode: FOCUS";
        }
        if (camera->getMode() == GameCamera::FOLLOW) {
            str += " mode: FOLLOW";
        }*/
        lab->setText(str);
        // <<<<

    }

    if (DEBUG_LOOP)printf("up draw\n");
    gs->update();
}

void GameControl::slotCheat1() {
    Bonus* newBonus = new BonusHeal;
    spawn(newBonus);
    gs->addModel(newBonus->getModel());
    bonuses.append(newBonus);
}

void GameControl::slotCheat2() {
    Bonus* newBonus = new BonusAmmo;
    spawn(newBonus);
    gs->addModel(newBonus->getModel());
    bonuses.append(newBonus);
}

void GameControl::slotCheat3() {
    Bonus* newBonus = new BonusLaser;
    spawn(newBonus);
    gs->addModel(newBonus->getModel());
    bonuses.append(newBonus);
}

void GameControl::slotCheat4() {
}

void GameControl::updatePlayers(int elapsed) {
    for (int i = 0; i < players.count(); i++) {
        Player *pl = players.at(i);

        if (pl->isDead()) {
            // dead
            if (pl->getTimer() > 2000) {
                spawn(pl);
                pl->reset();
            }
        } else {
            // alive
            if ((i && set->solo) || pl == sputnik) pl->ai->update(&players, &mapObjects);
            // movements
            if ((!((i && set->solo) || pl == sputnik) && cont->isLeft(i))
                    || (i && (set->solo || pl == sputnik) && pl->ai->isLeft())) {
                pl->steerLeft(elapsed);
            }
            if ((!((i && set->solo) || pl == sputnik) && cont->isRight(i))
                    || (i && (set->solo || pl == sputnik) && pl->ai->isRight())) {
                pl->steerRight(elapsed);
            }
            if ((!((i && set->solo) || pl == sputnik) && cont->isUp(i))
                    || (i && (set->solo || pl == sputnik) && pl->ai->isUp())) {
                pl->thrust(elapsed);
            }
            if ((!((i && set->solo) || pl == sputnik) && cont->isDown(i))
                    || (i && (set->solo || pl == sputnik) && pl->ai->isDown())) {
                pl->breaks(elapsed);
            }
            // shooting
            if (((!((i && set->solo) || pl == sputnik) && cont->isShoot(i))
                    || (i && (set->solo || pl == sputnik) && pl->ai->isShoot())) && players.at(i)->isWeaponReady()) {
                shoot(i);
            }
        }

        // move and check map bounds
        pl->update(elapsed);
        if (pl->getX() > 150)pl->setX(-150);
        if (pl->getX()<-150)pl->setX(150);
        if (pl->getY() > 100)pl->setY(-100);
        if (pl->getY()<-100)pl->setY(100);


        // portrait update
        if (pl != sputnik) {
            gs->getPortrait(i)->setHP(pl->getHp100());
            gs->getPortrait(i)->setShield(pl->getShield100());
            gs->getPortrait(i)->setAmmo(pl->getWeapon()->getAmmo100());
            gs->getPortrait(i)->setScore(pl->getScore());
        }

    }
}

void GameControl::updateObjects(int elapsed) {
    // update map objects
    for (int i = 0; i < mapObjects.count(); i++) {
        MapObject* m = mapObjects.at(i);
        m->update(elapsed);
    }
}

void GameControl::updateBonuses(int elapsed) {
    if (!set->modeNoBonus && bonusTime.elapsed() > 5000) {
        bonusTime.restart();
        // spawn bonus with chance
        if (rand() % 100 > 50) { // 50%
            int type = rand() % 3;

            Bonus* newBonus = NULL;
            switch (type) {
                case 0:
                    newBonus = new BonusHeal;
                    break;
                case 1:
                    newBonus = new BonusAmmo;
                    break;
                case 2:
                    newBonus = new BonusLaser;
                    break;
            }
            //printf("bon spawn\n");
            spawn(newBonus);
            //printf("bon add mod\n");
            gs->addModel(newBonus->getModel());
            //printf("bon append\n");
            bonuses.append(newBonus);
        }
    }

    //printf("bon for\n");
    for (int i = 0; i < bonuses.count(); i++) {
        Bonus* bon = bonuses.at(i);
        bon->update(elapsed);
        if (bon->isExpired()) {
            gs->removeModel(bon->getModel());
            bonuses.remove(i);
            i--;
            delete bon;
        }
    }
}

void GameControl::updateCollisions(int elapsed) {
    // players x players
    for (int a = 0; a < players.count() - 1; a++) {
        Player* ap = players.at(a);
        for (int b = 0; b < players.count(); b++) {
            if (a == b) continue;
            Player* bp = players.at(b);
            if (ap->collision(bp)) {
                // player collision, kill them both!
                ap->damage(99999);
                bp->damage(99999);

                ap->scoreUp(-1);
                bp->scoreUp(-1);

                if (bp == sputnik) {
                    ap->scoreUp(-1);
                }
            }
        }
    }

    // projectiles x players
    for (int a = 0; a < projectiles.count(); a++) {
        Projectile* ap = projectiles.at(a);
        for (int b = 0; b < players.count(); b++) {
            Player* bp = players.at(b);
            if (ap->collision(bp)) {
                if (ap->getOwner() == bp) continue;
                // player took hit
                bp->damage(ap->getDamage());
                ap->setExpired();
                if (bp->isDead()) {
                    // and died
                    Player* pl = dynamic_cast<Player*> (ap->getOwner());
                    pl->scoreUp(1);

                    if (bp == sputnik) {
                        pl->scoreUp(-3);
                    }
                }

            }
        }
    }

    // projectiles x mapobjects
    for (int a = 0; a < projectiles.count(); a++) {
        Projectile* ap = projectiles.at(a);
        for (int b = 0; b < mapObjects.count(); b++) {
            MapObject* bp = mapObjects.at(b);
            if (ap->collision(bp)) {
                // shooting rocks, remove bullet
                ap->setExpired();
            }

        }
    }

    // players x mapobjects
    for (int a = 0; a < players.count(); a++) {
        Player* ap = players.at(a);
        for (int b = 0; b < mapObjects.count(); b++) {
            MapObject* bp = mapObjects.at(b);
            if (ap->collision(bp)) {
                // that asteroid didnt move away =( Kill player
                ap->damage(99999);
                ap->scoreUp(-1);
            }

            if (ap->isDead() && ap->collisionHard(bp)) {
                ap->setSpeed(osg::Vec3(0, 0, 0));
            }
        }
    }

    // players x bonus

    for (int a = 0; a < players.count() - 1; a++) {
        Player* ap = players.at(a);
        for (int b = 0; b < bonuses.count(); b++) {
            Bonus* bp = bonuses.at(b);
            if (ap->collision(bp)) {
                // player picks up a bonus
                bp->apply(ap);
                gs->removeModel(bp->getModel());
            }
        }
    }


}

void GameControl::updateCamera(int elapsed) {
    // camera controls
    if (cont->isCamSwitch())camera->switchMode();
    if (cont->isCamZoom())camera->zoom(-1);
    if (cont->isCamUnzoom())camera->zoom(1);

    // @todo temp
    // camera movement set to player 4
    if (cont->isLeft(4)) {
        camera->focusMove(osg::Vec3(-1, 0, 0));
    }
    if (cont->isRight(4)) {
        camera->focusMove(osg::Vec3(1, 0, 0));
    }
    if (cont->isUp(4)) {
        camera->focusMove(osg::Vec3(0, 1, 0));
    }
    if (cont->isDown(4)) {
        camera->focusMove(osg::Vec3(0, -1, 0));
    }

    // update camera
    camera->update();
}

void GameControl::updateProjectiles(int elapsed) {
    // update projectile positions
    for (int i = 0; i < projectiles.count(); i++) {
        Projectile* p = projectiles.at(i);
        p->update(elapsed);
        if (p->getX() > 150) p->setX(-150);
        if (p->getX()<-150) p->setX(150);
        if (p->getY() > 100) p->setY(-100);
        if (p->getY()<-100) p->setY(100);

        if (p->isExpired()) {
            // remove if expired
            gs->removeModel(p->getModel());
            projectiles.remove(i);
            i--;
            //delete p;
        }
    }
}

void GameControl::shoot(int player) {
    Projectile* pr;
    Projectile* pr2;
    Player *pl = players.at(player);
    if (!pl->isLaser()) {
        // shoot normal projectiles
        pl->shoot(pr, pr2);
        gs->addModel(pr->getModel());
        projectiles.append(pr);
        if (pr2 != NULL) {
            gs->addModel(pr2->getModel());
            projectiles.append(pr2);
        }
    } else {
        // laser, have to find collision

        pl->shootLaser();
        for (int i = 1; i <= 2; i++) {
            osg::Vec3 point1;
            osg::Vec3 point2;
            point1 = pl->getShootPoint(i);

            double minDistance = 1000;

            double rad = osg::DegreesToRadians(pl->getYRotation());
            point2 = osg::Vec3(minDistance * cos(rad), minDistance * sin(rad), 0);

            // vs objects

            for (int b = 0; b < mapObjects.count(); b++) {
                MapObject* ob = mapObjects.at(b);
                osg::Vec3 objPoint = ob->getPosition() - point1;
                double distance = pl->distance(ob);
                double ata = osg::RadiansToDegrees(atan2(objPoint.x(), objPoint.y()));
                ata = 360 + 90 - ata;
                while (ata > 360)ata -= 360;
                double angle = ata - pl->getYRotation();
                if (abs(angle) > 90) continue;

                double col = distance * sin(osg::DegreesToRadians(angle));

                if (abs(col) < ob->getCollisonRadius()) {
                    // collision with object
                    if (distance < minDistance) {
                        minDistance = distance;
                        double dis = distance * cos(osg::DegreesToRadians(angle));
                        point2 = osg::Vec3(dis * cos(rad), dis * sin(rad), 0);
                    }
                }
            }
            // vs player

            Player* playerHit = NULL;
            for (int a = 0; a < players.count(); a++) {
                Player* ap = players.at(a);
                if (ap == pl) continue;
                if (ap->isCollisionImmune())continue;

                double distance = pl->distance(ap);
                //double angle = pl->angleTo(ap);
                osg::Vec3 objPoint = ap->getPosition() - point1;
                double ata = osg::RadiansToDegrees(atan2(objPoint.x(), objPoint.y()));
                ata = 360 + 90 - ata;
                while (ata > 360)ata -= 360;
                double angle = ata - pl->getYRotation();
                if (abs(angle) > 90) continue;

                double col = distance * sin(osg::DegreesToRadians(angle));
                if (abs(col) < ap->getCollisonRadius()) {
                    // collision with object
                    if (distance < minDistance) {
                        playerHit = ap;
                        minDistance = distance;
                        double dis = distance * cos(osg::DegreesToRadians(angle));
                        point2 = osg::Vec3(dis * cos(rad), dis * sin(rad), 0);
                    }
                }
            }
            point2 += point1;

            pr = new Beam(point1, point2);
            projectiles.append(pr);
            gs->addModel(pr->getModel());
            pr->setSpeed(pl->getSpeed());

            if (playerHit) {
                playerHit->damage(pr->getDamage());

                if (playerHit->isDead()) {
                    // and died
                    pl->scoreUp(1);
                    if (playerHit == sputnik) {
                        pl->scoreUp(-3);
                    }
                }

            }
        }
    }

}

void GameControl::spawn(GameObject *p) {
    bool ok = false;
    while (!ok) {
        // spawn players at random position
        ok = true;
        int x = (rand() % 250) - 125;
        int y = (rand() % 150) - 75;
        p->setX(x);
        p->setY(y);

        // and check collisions

        // obj x player
        for (int b = 0; b < players.count(); b++) {
            Player* bp = players.at(b);
            if (p == bp) continue;
            if (p->collisionHard(bp)) {
                ok = false;
            }
        }

        // obj x mapobjects
        for (int b = 0; b < mapObjects.count(); b++) {
            MapObject* bp = mapObjects.at(b);
            if (bp->collisionHard(p->getPosition(), p->getCollisonRadius() + 3)) {
                ok = false;
            }
        }
    }
}

void GameControl::addMapObject(MapObject *obj) {
    mapObjects.append(obj);
    gs->addModel(obj->getModel());
}
