/*****************************************************************************\
 *
 * Author: Michal Tth (xtothm00 _AT stud.fit.vutbr.cz)
 *
 * ----------------------------------------------------------------------------
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, author will kindly welcome
 * if you give him credit and keep his names with their source code
 * and inform him about any succesful project using this code.
 *
\*****************************************************************************/
#include "connection.h"

Connection::Connection(QObject *parent) :
QObject(parent) {
    socketUDP = new QUdpSocket();
    socketTCP = new QTcpSocket();

    sent = 0;
    sentps = 0;
    recieved = 0;
    recievedps = 0;
    time.start();
}

Connection::~Connection() {
    delete socketUDP;
    delete socketTCP;
}

QString* Connection::init(QString ip, int port) {
    server.addr.setAddress(ip);
    server.port = port;
    socketTCP->connectToHost(server.addr, port);
    if (!socketTCP->waitForConnected(2000)) {
        QString s;
        printf("%s\n", socketTCP->errorString().toStdString().c_str());
        return new QString("TCP_CONNECT_FAIL");
    }
    if (!socketUDP->bind(QHostAddress::Any, socketTCP->localPort())) {
        return new QString("UDP_CONNECT_FAIL");
    }

    connect(socketTCP, SIGNAL(readyRead()), this, SLOT(slotTCPReady()));
    connect(socketUDP, SIGNAL(readyRead()), this, SLOT(slotUDPReady()));

    connect(socketTCP, SIGNAL(disconnected()), this, SLOT(disconnected()));
    return NULL;
}

void Connection::sendFast(Packet *p) {

    if (USE_UDP) {
        QByteArray d = p->serialize();
        sendUDP(d, socketUDP, server.addr, server.port);
    } else {
        send(p);
    }
}

void Connection::send(Packet *p) {

    if (USE_TCP) {
        QByteArray d = p->serialize();
        sendTCP(d, socketTCP);
    } else {
        sendFast(p);
    }
}

void Connection::sendTCP(QByteArray &p, QTcpSocket *soc) {
    //printf("send TCP\n");
    if (soc == NULL)return;
    sent += p.length();
    soc->write(p);
}

void Connection::sendUDP(QByteArray &p, QUdpSocket *soc, QHostAddress addr, int port) {
    //printf("send UDP\n");
    sent += p.length();
    soc->writeDatagram(p, addr, port);
}

void Connection::slotTCPReady() {
    readTCP(socketTCP);
    emit packetReady();
}

void Connection::readTCP(QTcpSocket *soc) {
    //printf("read TCP\n");
    uint length;
    while ((soc->read((char *) &length, sizeof (uint))) == sizeof (uint)) {
        QByteArray packet;
        packet.resize(length);

        uint re = 0;
        while (re != length) {
            re += soc->read(packet.data() + re, length - re);
        }
        packet.prepend((char*) &length, sizeof (uint));
        processPacket(packet);
    }
}

void Connection::slotUDPReady() {
    readUDP(socketUDP);
    emit packetReady();
}

void Connection::readUDP(QUdpSocket *soc) {
    //printf("read udp");
    while (soc->hasPendingDatagrams()) {
        QByteArray datagram;
        datagram.resize(soc->pendingDatagramSize());
        QHostAddress sender;
        quint16 senderPort;
        soc->readDatagram(datagram.data(), datagram.size(),
                &sender, &senderPort);

        if (sender != server.addr && senderPort != server.port) {
            //printf("packet dropped\n");
            continue;
        }
        processPacket(datagram);
    }
}

Packet* Connection::read() {
    if (packetQueue.empty()) return NULL;
    Packet* p = packetQueue.first();
    packetQueue.pop_front();
    return p;
}

void Connection::processPacket(QByteArray &packet) {
    recieved += packet.length();
    int elapsed = time.elapsed();
    if (elapsed > 1000) {
        sentps = sent * 1000 / elapsed;
        sent = 0;
        recievedps = recieved * 1000 / elapsed;
        recieved = 0;
        time.restart();
    }


    Packet p(packet);
    //printf("packet type= %d\n",p.type);
    switch (p.type) {
        case Packet::GAME_FULL:
        case Packet::GAME_START:
            packetQueue.append(new Packet(packet));
            break;
        case Packet::CHAT:
        case Packet::NAME_SELECT:
            packetQueue.append(new ChatPacket(packet));
            break;
        case Packet::ASIGN_NUMBER:
        case Packet::MODEL_SELECT:
        case Packet::PLAYER_JOINED:
        case Packet::SHOOT:
        case Packet::WANT_SHOOT:
        case Packet::DISCONNECTED:
        case Packet::DAMAGE:
        case Packet::KILL:
        case Packet::PING:
        case Packet::PING_REPLY:
        case Packet::BONUS_REMOVE:
            packetQueue.append(new IntPacket(packet));
            break;
        case Packet::CHOOSER_STATE:
            packetQueue.append(new ChooserPacket(packet));
            break;
        case Packet::COLLISION:
            packetQueue.append(new CollisionPacket(packet));
            break;
        case Packet::SPAWN:
        case Packet::PLAYER_POSITION:
            packetQueue.append(new PositionPacket(packet));
            break;
        case Packet::SCORE:
            packetQueue.append(new ScorePacket(packet));
            break;
        case Packet::BONUS_SPAWN:
            packetQueue.append(new BonusPacket(packet));
            break;
        default:
            break;
    }
}

void Connection::disconnected() {
    printf("socket disconnected \n");
    socketTCP->deleteLater();
    socketTCP = NULL;
    packetQueue.append(new IntPacket(Packet::DISCONNECTED, 0, 0));
    emit packetReady();
}
